/*****************************************************
**********     by Luc Bissonnette     ****************
**********     Universite Laval       ****************
**********     CIRPEE                 ****************
**********     July 11th 2014         ****************
*****************************************************/                             

capture program drop splinesBBK

program define splinesBBK


 version 10.1
 syntax newvarlist (generate), QName(string) QList(numlist) Knots(varlist) Prob(varlist) [Hyman(integer 1) EPSilon(real 0.001) RANDName(string) RANDDraws(integer 0)]
 tokenize `varlist' 

// scalar `ithisobs'  = 1
 local totobs = _N

****** Creating the variable with the expected values ********

* Default name for the random draws 
if "`randname'" == "" local randname "ran_"
foreach quant in `qlist'{
  local perc = `quant'*100
  qui gen `qname'`perc' = .
}

***** Creating variables for random draws *****
forvalues i = 1/`randdraws'{    //if 0 (default), then no iterations.
  qui gen `randname'`i' = .
}


//no di "hello `knots' `prob'"




****** Printing information on screen ********

if(`hyman'){
 di "Applying Hyman filter"
 qui gen hyman = 0
}
else{
 di "Not applying Hyman filter"
}





tempvar id
qui gen `id' = _n

tempvar b
qui gen `b' = 0

tempvar nb_ok_prob
qui gen `nb_ok_prob' = 0
foreach var in `prob'{
qui replace `nb_ok_prob' = `nb_ok_prob' + !missing(`var')
}



/******* The code is executed quitetly. *******/
qui{
forvalues ithisobs = 1/`totobs'{

if(`nb_ok_prob'[`ithisobs'] > 1){



******* Creating matrices for knots and prob *********
 mat TEMP = . 

foreach var in `knots'{
if(!missing(`var'[`ithisobs' ])){
 mat TEMP = TEMP\(`var'[`ithisobs' ])
 mat mknots = TEMP[2...,1]  
 }
}

 mat TEMP = . 

foreach var in `prob'{
if(!missing(`var'[`ithisobs' ])){
 mat TEMP = TEMP\(`var'[`ithisobs' ]) 
 mat mprob = TEMP[2...,1]  
 }
}





****** We want to create a system Ax = b *******

******* Creating the system of equations A-transpose. One column is an equation **********


capture drop SYS*

local nobs = _N
local leq = rowsof(mprob) - 1      //Number of polynomials (eq for equations)
local lparam = 4*`leq'             //Number of parameters. So far, only cubic splines.


if (`lparam' > `nobs'){
  set obs `lparam'
}

//Initializing the vector b to 0
replace `b' = .
replace `b' = 0 if _n <= `lparam'


//Initialize a matrices of 0 IN THE DATASET
forvalues i = 1/`lparam'{
  capture gen SYS`i' = 0 if _n <= `lparam' //Not temporary var, to make it easier to create matrices...
}

//The first conditions are for the first knot of the interval

forvalues i = 1/`leq'{
   
  replace `b' = mprob[`i',1] in `i'

  forvalues j=0/3{
   local invalue = (4*(`i'-1) + `j' + 1)
   replace SYS`i' = mknots[`i',1]^`j' in `invalue' 

  } 
}



//The next conditions are for the last knot of the interval

forvalues i = 1/`leq'{
  local constr = `i' + `leq'               //Constr used to know which constraint is being handled.
  replace `b' = mprob[`i'+1,1] in `constr'

  forvalues j=0/3{
   local invalue = (4*(`i'-1) + `j' + 1)
   replace SYS`constr' = mknots[`i'+1,1]^`j' in `invalue'

  } 
}



//The next conditions are the first derivatives in the interior knots

local lint = `leq' - 1   //Counts the number of interior nodes.

forvalues i = 1/`lint'{
  local constr = `i' + 2*`leq' 

  forvalues j=1/3{
   local invalue = (4*(`i'-1) + `j' + 1)
   replace SYS`constr' = `j'*mknots[`i'+1,1]^(`j'-1) in `invalue'

   local invalue = (4*(`i') + `j' + 1)
   replace SYS`constr' = -`j'*mknots[`i'+1,1]^(`j'-1) in `invalue'

  } 
}

//The next conditions are the second derivatives in the interior knots
forvalues i = 1/`lint'{
  local constr = `i' + 2*`leq' + `lint'

  forvalues j=2/3{
   local invalue = (4*(`i'-1) + `j' + 1)
   replace SYS`constr' = (`j'-1)*`j'*mknots[`i'+1,1]^(`j'-2) in `invalue'

   local invalue = (4*(`i') + `j' + 1)
   replace SYS`constr' = -(`j'-1)*`j'*mknots[`i'+1,1]^(`j'-2) in `invalue'

  } 
}

//Finally, the terminal conditions

  local constr = `lparam' - 1 
  forvalues j=2/3{
   local invalue = `j' + 1
   replace SYS`constr' = (`j'-1)*`j'*mknots[1,1]^(`j'-2) in `invalue'
  }
  
  
 forvalues j=2/3{
   local invalue = `lparam'-3 + `j'
   replace SYS`lparam' = (`j'-1)*`j'*mknots[`leq'+1,1]^(`j'-2) in `invalue'
  }


//We store the matrices in the right system.

mkmat SYS*, matrix(A) nomissing
mkmat `b', matrix(b) nomissing

mat x = inv(A')*b     //Note the transpose for A. x contains the spline parameters in a column vector.
mat list x



/****** Hyman filter *****/
if(`hyman'){
 forvalues i = 1/`lint'{

  scalar Sl = (mprob[`i'+1,1]-mprob[`i',1]) /(mknots[`i'+1,1] - mknots[`i',1])
  scalar Su = (mprob[`i'+2,1]-mprob[`i'+1,1])/(mknots[`i'+2,1] - mknots[`i'+1,1])
  //di 3*min(Sl,Su)

  scalar fhat = x[4*(`i')+2,1] + 2*x[4*(`i')+3,1]*mknots[`i'+1,1] ///
                + 3*x[4*(`i')+4,1]*mknots[`i'+1,1]^2
  
//  no di mknots[`i'+1,1], x[4*(`i')+2,1],  2*x[4*(`i')+3,1]*mknots[`i'+1,1] , 3*x[4*(`i')+4,1]*mknots[`i'+1,1]^2, fhat
//  no mat list x
//    no di fhat, Sl, Su,3*min(Sl,Su), 0<fhat, fhat<3*min(Sl,Su)

// If the derivative is smaller than 1
  if (fhat<0){
   replace hyman = hyman in `ithisobs'
   local constr = `i' + 2*`leq' + `lint'

   //Reset the relevant constraint:
   replace SYS`constr' = 0 if !missing(SYS`constr')   

   forvalues j=1/3{
     local invalue = (4*(`i'-1) + `j' + 1)
     replace SYS`constr' = `j'*mknots[`i'+1,1]^(`j'-1) in `invalue'
  }
 }

//If the derivative is larger than 3*min(Sl,Su)
  else if(fhat>3*min(Sl,Su)){
   replace hyman = 1 in `ithisobs'
   local constr = `i' + 2*`leq' + `lint'
    
   //Reset the relevant constraint:
   replace SYS`constr' = 0 if !missing(SYS`constr')   

   forvalues j=1/3{
     local invalue = (4*(`i'-1) + `j' + 1)
     replace SYS`constr' = `j'*mknots[`i'+1,1]^(`j'-1) in `invalue'
     replace `b' = 3*min(Sl,Su) in `constr' 
    }
   }
} 

mkmat SYS*, matrix(A) nomissing
mkmat `b', matrix(b) nomissing

mat x = inv(A')*b     //Note the transpose for A. x contains the spline parameters in a column vector.
//mat list x



} //end of filter if
/***** End of the filter ******/





/***** Expected value *****/
scalar expect = mprob[1,1]*mknots[1,1] + (1-mprob[`leq'+1,1])*mknots[`leq'+1,1]  //Censoring
forvalues i = 1/`leq'{
  scalar expect = expect ///
                + x[4*(`i'-1)+2,1]*(mknots[`i'+1,1]^2-mknots[`i',1]^2)/2     ///
                + x[4*(`i'-1)+3,1]*2*(mknots[`i'+1,1]^3-mknots[`i',1]^3)/3   ///
                + x[4*(`i'-1)+4,1]*3*(mknots[`i'+1,1]^4-mknots[`i',1]^4)/4 
}


replace `1' = expect in `ithisobs' 
  

/***** Quantiles *****/

foreach quant in `qlist'{
  local perc = `quant'*100
  scalar loc = 0 

 if(mprob[1,1] >= `quant') {
    replace `qname'`perc'  = `=mknots[1,1]' in `ithisobs'  
   }
 else if(mprob[`leq'+1,1] <= `quant') {
    replace `qname'`perc'  = `=mknots[`leq'+1,1]' in `ithisobs'  
   }

 else {
  forvalues i = 1/`leq'{
     if((mprob[`i',1] <= `quant') & (mprob[`i'+1,1] >= `quant')) scalar loc = `i'
  }
  scalar val = -1
  scalar j = mknots[loc,1]

  while val < `quant' {
    scalar val = x[4*(loc-1)+1,1] + x[4*(loc-1)+2,1]*j + x[4*(loc-1)+3,1]*j^2 + x[4*(loc-1)+4,1]*j^3
    if(val >= `quant') replace `qname'`perc' = j in `ithisobs' 
    scalar j = j + `epsilon'
  } 
 } 

}

/****** Random draws ********/
forvalues i = 1/`randdraws'{    //if 0 (default), then no iterations.
  local draw = uniform()

// Can't seem to make it work if I do what I did above...
    local name = "`randname'`i'"
    //no di "`name'  " `draw'   
  
  scalar loc = 0 
  
    
  if(mprob[1,1] >= `draw') {
    replace `name'  = `=mknots[1,1]' in `ithisobs'  
   }
 else if(mprob[`leq'+1,1] <= `draw') {
    replace `name'  = `=mknots[`leq',1]' in `ithisobs'  
   }

 else {
  forvalues i = 1/`leq'{
     if((mprob[`i',1] <= `draw') & (mprob[`i'+1,1] >= `draw')) scalar loc = `i'
  }
  scalar val = -1
  scalar j = mknots[loc,1]


  while val < `draw' {
    scalar val = x[4*(loc-1)+1,1] + x[4*(loc-1)+2,1]*j + x[4*(loc-1)+3,1]*j^2 + x[4*(loc-1)+4,1]*j^3
    if(val >= `draw') replace `name'`i' = j in `ithisobs' 
    scalar j = j + `epsilon'
  }  
 }  
    
}    
   
  

 


     
	 
//capture drop `b'	 
	 
	 } //end of OK conditions




   } // end of obs loop
 } // end of qui

drop SYS* //Not temporary var, to make it easier to create matrices...
qui drop if missing(`id') //Removes the additional lines.


end



